"""
PyZITADELLE is a quick asynchronous framework for testing python applications
Copyright (C) 2024	Alexeev Bronislav (C) 2024

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
USA
"""

import requests
from rich import print

__version__ = "0.2.1"


def check_for_update():
    """
    Check for update in pypi
    """
    try:
        response = requests.get("https://pypi.org/pypi/pyzitadelle/json").json()

        latest_version = response["info"]["version"]

        latest_digits = [int(n) for n in latest_version.split(".")]
        current_digits = [int(n) for n in __version__.split(".")]

        if sum(latest_digits) > sum(current_digits):
            message = f"New version of library pyzitadelle available: {latest_version}"

            print(
                f"[red]{'#' * (len(message) + 4)}\n#[/red][bold yellow] {message} [/bold yellow][red]#\n{'#' * (len(message) + 4)}[/red]\n"
            )
        elif sum(latest_digits) < sum(current_digits):
            print(
                f"[yellow]You use [bold]UNSTABLE[/bold] branch of pyzitadelle. Stable version: {latest_version}, your version: {__version__}[/yellow]\n"
            )
    except (requests.RequestException, KeyError):
        print(
            f"[dim]Version updates information not available. Your version: {__version__}[/dim]"
        )


check_for_update()

print(
    f"""[white]
                 _ __          __    ____
   ___ __ _____ (_) /____ ____/ /__ / / /__
  / _ \\/ // /_ // / __/ _ `/ _	 / -_) // -_)
 / .__/\\_, //__/_/\\__/\\_,_/\\_,_/\\__/_/_/\\__/	  [bold]v{__version__}[/bold]
/_/	  /___/[/white]
	"""
)



from functools import partial, wraps
from time import time
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union

from pyzitadelle.exceptions import TestError
from pyzitadelle.reporter import print_header, print_results_table
from pyzitadelle.sessions import Runner
from pyzitadelle.standard import (
	Argument,
	CollectionMetadata,
	Each,
	ExpectFailMarkup,
	Fixture,
	SkipMarker,
)


def skip(
	func_or_reason: Union[str, Callable, None] = None,
	*,
	reason: Optional[str] = None,
	when: Union[bool, Callable] = True,
):
	"""
	Decorator which can be used to optionally skip tests.
	"""
	if func_or_reason is None:
		return partial(skip, reason=reason, when=when)

	if isinstance(func_or_reason, str):
		return partial(skip, reason=func_or_reason, when=when)

	func = func_or_reason

	marker = SkipMarker(reason=reason, when=when)

	if hasattr(func, "ward_meta"):
		func.pztdmeta.marker = marker  # type: ignore[attr-defined]
	else:
		func.pztdmeta = CollectionMetadata(marker=marker)  # type: ignore[attr-defined]

	@wraps(func)
	def wrapper(*args, **kwargs):
		return func(*args, **kwargs)

	return wrapper


def expectfail(
	func_or_reason: Union[str, Callable, None] = None,
	*,
	reason: Optional[str] = None,
	when: Union[bool, Callable] = True,
):
	if func_or_reason is None:
		return partial(expectfail, reason=reason, when=when)

	if isinstance(func_or_reason, str):
		return partial(expectfail, reason=func_or_reason, when=when)

	func = func_or_reason
	marker = ExpectFailMarkup(reason=reason, when=when)

	if hasattr(func, "pztdmeta"):
		func.pztdmeta.marker = marker  # type: ignore[attr-defined]
	else:
		func.pztdmeta = CollectionMetadata(marker=marker)  # type: ignore[attr-defined]

	@wraps(func)
	def wrapper(*args, **kwargs):
		return func(*args, **kwargs)

	return wrapper


class BaseTestCase:
	"""
	This class describes a base test case.
	"""

	def __init__(self, label: str = "TestCase"):
		"""
		Constructs a new instance.

		:param		label:	The label
		:type		label:	str
		"""
		self.label: str = label

		self.warnings: int = 0
		self.tags: List[str] = []
		self.fixtures: Dict[str, Fixture] = []
		self.skipped: int = 0
		self.errors: int = 0
		self.passed: int = 0

		self.tests: Dict[str, Union[Callable, Awaitable]] = {}


class TestCase(BaseTestCase):
	"""
	This class describes a test case.
	"""

	def __init__(self, label: str = "TestCase"):
		"""
		Constructs a new instance.

		:param		label:	The label
		:type		label:	str
		"""
		super().__init__(label)

	def fixture(self):
		def wrapper(
			func: Union[Awaitable, Callable], *args, **kwargs
		) -> Union[Awaitable, Callable]:
			if not hasattr(func, "pztdmeta"):
				func.pztdmeta = CollectionMetadata(is_fixture=True)
			else:
				func.pztdmeta.is_fixture = True

			self.fixtures[func.__name__] = Fixture(handler=func)

			print('\nFixtures in DEV\n')

			return func(*args, **kwargs)

		return wrapper

	def test(
		self,
		comment: str = None,
		tags: List[str] = [],
		count_of_launchs: int = 1,
		arguments: Tuple[Argument] = (),
	) -> Callable:
		"""
		Add test to environment

		:param		comment:		   The comment
		:type		comment:		   str
		:param		tags:			   The tags
		:type		tags:			   Array
		:param		count_of_launchs:  The count of launchs
		:type		count_of_launchs:  int
		:param		skip_test:		   The skip test
		:type		skip_test:		   bool
		:param		arguments:		   The arguments
		:type		arguments:		   Tuple[Argument]

		:returns:	wrapper
		:rtype:		Callable
		"""

		def wrapper(
			func: Union[Awaitable, Callable], *args, **kwargs
		) -> Union[Awaitable, Callable]:
			"""
			Wrapper for @test decorator

			:param		func:	 The function
			:type		func:	 Union[Awaitable, Callable]
			:param		args:	 The arguments
			:type		args:	 list
			:param		kwargs:	 The keywords arguments
			:type		kwargs:	 dictionary

			:returns:	function
			:rtype:		Union[Awaitable, Callable]
			"""
			if not hasattr(func, "pztdmeta"):
				func.pztdmeta = CollectionMetadata(
					comment=comment.format(**kwargs) if comment is not None else None,
					tags=tags,
					arguments=arguments,
					count_of_launchs=count_of_launchs,
				)
			else:
				func.pztdmeta.comment = (
					comment.format(**kwargs) if comment is not None else None
				)
				func.pztdmeta.tags = tags
				func.pztdmeta.arguments = arguments
				func.pztdmeta.count_of_launchs = count_of_launchs

			self.tags = list(set(self.tags + tags))

			self.tests[func.__name__] = func
			return func

		return wrapper

	def run(self, tags: Optional[List[str]] = []):
		"""
		Run testing
		"""
		runner = Runner(self.tests, self)

		start = time()

		runner.launch_test_chain(tags=tags)

		end = time()
		total = end - start

		print_header(
			f"[cyan]{len(self.tests)} tests runned {round(total, 2)}s[/cyan]",
			plus_len=15,
		)

		print_results_table(
			len(self.tests), self.passed, self.warnings, self.errors, self.skipped
		)


def expect(lhs: Any, rhs: Any, message: str) -> bool:
	"""
	Expect lhs and rhs with message

	:param		lhs:		The left hand side
	:type		lhs:		Any
	:param		rhs:		The right hand side
	:type		rhs:		Any
	:param		message:	The message
	:type		message:	str

	:returns:	true is equals, raise error otherwise
	:rtype:		bool

	:raises		TestError:	lhs and rhs is not equals
	"""
	if lhs == rhs:
		return True
	else:
		raise TestError(message)


def each(*args):
	return Each(args)



class TestError(Exception):
	def __init__(self, *args):
		"""
		Constructs a new instance.

		:param		args:  The arguments
		:type		args:  list
		"""
		if args:
			self.message = args[0]
		else:
			self.message = None

	def get_explanation(self) -> str:
		"""
		Gets the explanation.

		:returns:	The explanation.
		:rtype:		str
		"""
		return f"Message: {self.message if self.message else 'missing'}"

	def __str__(self):
		"""
		Returns a string representation of the object.

		:returns:	String representation of the object.
		:rtype:		str
		"""
		return f"TestError has been raised. {self.get_explanation()}"


class SkippedTestException(TestError):
	def __init__(self, *args):
		"""
		Constructs a new instance.

		:param		args:  The arguments
		:type		args:  list
		"""
		if args:
			self.message = args[0]
		else:
			self.message = "SkippedTest"

	def __str__(self):
		"""
		Returns a string representation of the object.

		:returns:	String representation of the object.
		:rtype:		str
		"""
		return f"{self.message}"


class TestValidationError(TestError):
	def __str__(self):
		"""
		Returns a string representation of the object.

		:returns:	String representation of the object.
		:rtype:		str
		"""
		return f"TestValidationError has been raised. {self.get_explanation()}"


class FixtureError(TestError):
	def __str__(self):
		"""
		Returns a string representation of the object.

		:returns:	String representation of the object.
		:rtype:		str
		"""
		return f"FixtureError has been raised. {self.get_explanation()}"



import asyncio
import inspect
import traceback
from typing import Any, Awaitable, Callable, List, Union

from pyzitadelle.exceptions import SkippedTestException, TestError
from pyzitadelle.reporter import print_header, print_platform, print_test_result
from pyzitadelle.standard import ExpectFailMarkup, SkipMarker


class Runner:
	"""
	This class describes a runner session.
	"""

	def __init__(self, tests: int, testcase: object):
		"""
		Constructs a new instance.

		:param		tests:	   The tests
		:type		tests:	   int
		:param		testcase:  The testcase
		:type		testcase:  TestCase
		"""
		self.tests = tests
		self.tests_count = len(self.tests)
		self.testcase = testcase

	def _print_prelude(self):
		"""
		Prints a prelude.
		"""
		print_header("runner session starts")

		print_platform(self.tests_count)

	def _run_testinfo(self, test: Union[Callable, Awaitable], *args, **kwargs) -> Any:
		"""
		Run test with args

		:param		test:	 The test
		:type		test:	 TestInfo
		:param		args:	 The arguments
		:type		args:	 list
		:param		kwargs:	 The keywords arguments
		:type		kwargs:	 dictionary

		:returns:	function result
		:rtype:		Any
		"""
		if inspect.iscoroutinefunction(test):
			result = asyncio.run(test(*args, **kwargs))
		else:
			result = test(*args, **kwargs)

		return result

	def _run_test_cycle(self, test_name: str, test: Union[Awaitable, Callable]) -> Any:
		"""
		Run test launch cycle

		:param		test_name:	The test name
		:type		test_name:	str
		:param		test:		The test
		:type		test:		TestInfo

		:returns:	function result
		:rtype:		Any
		"""
		for n in range(test.pztdmeta.count_of_launchs):
			if test.pztdmeta.arguments:
				for argument in test.pztdmeta.arguments:
					result = self._run_testinfo(test, *argument.args, **argument.kwargs)
			else:
				result = self._run_testinfo(test)

		return result

	def _check_warnings(self, result: Any, results: list, percent: int, test_name: str):
		"""
		Check warnings in test

		:param		result:		The result
		:type		result:		Any
		:param		results:	The results
		:type		results:	list
		:param		percent:	The percent
		:type		percent:	int
		:param		test_name:	The test name
		:type		test_name:	str
		"""
		if len(results) > 0 and results[-1] == result and result is not None:
			print_test_result(
				percent,
				test_name,
				status="warning",
				output=f"Last result is equals current result ({results[-1]} == {result})",
			)
			self.testcase.warnings += 1
			self.testcase.passed += 1

	def _processing_tests_execution(
		self,
		tags: List[str],
		test_num: int,
		test_name: str,
		test: Union[Awaitable, Callable],
	):
		"""
		Processing tests execution

		:param		tags:				   The tags
		:type		tags:				   List[str]
		:param		test_num:			   The test number
		:type		test_num:			   int
		:param		test_name:			   The test name
		:type		test_name:			   str
		:param		test:				   The test
		:type		test:				   TestInfo

		:raises		SkippedTestException:  skip test
		"""
		percent = int((test_num / self.tests_count) * 100)
		results = []

		lines = inspect.getsourcelines(test)[1]
		test_name = f"{test_name}:[line {lines}]"

		try:
			if tags and list(set(tags) & set(test.tags)):
				raise SkippedTestException()
			elif isinstance(test.pztdmeta.marker, SkipMarker):
				marker = test.pztdmeta.marker

				if marker.when:
					raise SkippedTestException(
						marker.reason if marker.reason else "SkippedTest"
					)
			elif isinstance(test.pztdmeta.marker, ExpectFailMarkup):
				marker = test.pztdmeta.marker

			result = self._run_test_cycle(test_name, test)

			self._check_warnings(result, results, percent, test_name)

			results.append(result)
		except SkippedTestException as ex:
			self.testcase.skipped += 1
			print_test_result(
				percent,
				test_name,
				status="skip",
				postmessage=str(ex),
				comment=test.pztdmeta.comment,
			)
		except (AssertionError, TestError):
			self.testcase.errors += 1

			if isinstance(marker, ExpectFailMarkup):
				print_test_result(
					percent,
					test_name,
					status="error",
					output=f"{traceback.format_exc()}",
					postmessage=marker.reason if marker.reason else "XFAIL",
					comment=test.pztdmeta.comment,
				)
			else:
				print_test_result(
					percent,
					test_name,
					status="error",
					output=f"{traceback.format_exc()}",
					comment=test.pztdmeta.comment,
				)
		else:
			self.testcase.passed += 1

			print_test_result(percent, test_name, comment=test.pztdmeta.comment)

	def launch_test_chain(self, tags: List[str]):
		"""
		Launch test chain

		:raises		SkippedTestException:  skip test

		:param		tags:  The tags
		:type		tags:  List[str]
		"""
		for test_num, (test_name, test) in enumerate(self.tests.items(), start=1):
			self._processing_tests_execution(tags, test_num, test_name, test)



import platform
import shutil
from datetime import datetime
from typing import Any, Optional

from rich import box, print
from rich.console import Console
from rich.table import Table


def print_results_table(
	total: int, passed: int, warnings: int, errors: int, skipped: int
):
	table = Table(title="Tests Result", expand=True, box=box.ROUNDED)

	table.add_column("N", style="cyan")
	table.add_column("Tests encountered", style="cyan")
	table.add_column("Percent", style="cyan")

	passed_percent = int((passed / total) * 100)
	warnings_percent = int((warnings / total) * 100)
	errors_percent = int((errors / total) * 100)
	skipped_percent = int((skipped / total) * 100)

	table.add_row(str(total), "Total", "100%")
	table.add_row(
		str(passed), "Passed", f"{passed_percent}%", style="black bold on green"
	)
	table.add_row(
		str(warnings), "Warnings", f"{warnings_percent}%", style="black bold on yellow"
	)
	table.add_row(
		str(errors), "Errors", f"{errors_percent}%", style="black bold on red"
	)
	table.add_row(
		str(skipped), "Skipped", f"{skipped_percent}%", style="black bold on blue"
	)

	console = Console()
	console.print(table)


def print_header(label: str, plus_len: int = 0, style: str = "bold"):
	"""
	Prints a header.

	:param		label:	   The label
	:type		label:	   str
	:param		plus_len:  The plus length
	:type		plus_len:  int
	:param		style:	   The style
	:type		style:	   str
	"""
	width = shutil.get_terminal_size().columns - 2 + plus_len

	line = f" {label} ".center(width, "=")

	print(f"[{style}]{line}[/{style}]")


def print_platform(items: int):
	"""
	Prints a platform.

	:param		items:	The items
	:type		items:	int
	"""
	print(f"[white]platform: [reset]{platform.platform()}[/white]")
	print(f"[white]version: [reset]{platform.version()}[/white]")
	print(f"[white]release: [reset]{platform.release()}[/white]")
	print(f"[white]system: [reset]{platform.system()}[/white]")
	print(f"[white]python: [reset]{platform.python_version()}[/white]")
	print(f"[white bold]Collected {items} items[/white bold]\n")


def print_comment(comment: str):
	print(f"[dim]{comment}[/dim]")


def print_test_result(
	percent: str,
	label: str,
	status: Optional[str] = "success",
	output: Optional[Any] = None,
	postmessage: Optional[str] = "",
	comment: Optional[str] = None,
):
	"""
	Prints a test result.

	:param		percent:  The percent
	:type		percent:  str
	:param		label:	  The label
	:type		label:	  str
	:param		status:	  The status
	:type		status:	  str
	:param		output:	  The output
	:type		output:	  Any
	"""
	date = datetime.now().strftime("%d-%m-%Y %H:%M:%S")
	width = shutil.get_terminal_size().columns - 13 - len(date) - len(postmessage)

	if comment is not None:
		label = f"[dim]{label}[/dim] [white]{comment}[/white]"
		width += 26

	if status == "success":
		print(
			f"[black bold on green]PASS[/black bold on green] {date} [green]{label.ljust(width)}[/green][black on blue]{postmessage}[/black on blue] [dim green][{str(percent).rjust(3)}%][/dim green]"
		)
	elif status == "error":
		print(
			f"\n[black bold on red]ERR [/black bold on red] {date} [red]{label.ljust(width)}[/red][black on blue]{postmessage}[/black on blue] [dim red][{str(percent).rjust(3)}%][/dim red]"
		)
		print_header(f"ERROR: {label}", style="bold red")
		print(f"[red]{output}[/red]")
	elif status == "warning":
		print(
			f"[black bold on yellow]WARN[/black bold on yellow] {date} [yellow]{label.ljust(width)}[/yellow][black on blue]{postmessage}[/black on blue] [dim yellow][{str(percent).rjust(3)}%][/dim yellow]"
		)
		print(f"[yellow] > {output}[/yellow]\n")
	elif status == "skip":
		print(
			f"[black bold on blue]SKIP[/black bold on blue] {date} [blue]{label.ljust(width)}[/blue][black on blue]{postmessage}[/black on blue] [dim blue][{str(percent).rjust(3)}%][/dim blue]"
		)



from functools import wraps
from time import time
from typing import Callable

from rich import print


def async_debug_measurement(label: str = "measurement") -> Callable:
	def decorator(func):
		@wraps(func)
		async def wrapper(*args, **kwargs):
			start = time()
			result = await func(*args, **kwargs)
			end = time()

			total = round(end - start, 9)

			print(
				"[bold dim]{}[/bold dim] : {}".format(
					f"{func.__name__} | {label}".ljust(len(label) * 2), total
				)
			)

			return result


def debug_measurement(label: str = "measurement"):
	def decorator(func):
		@wraps(func)
		def wrapper(*args, **kwargs):
			start = time()
			result = func(*args, **kwargs)
			end = time()

			total = round(end - start, 9)

			print(
				"[bold dim]{}[/bold dim] : {}".format(
					f"{func.__name__} | {label}".ljust(len(label) * 2), total
				)
			)

			return result

		return wrapper

	return decorator



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import (
	Any,
	AsyncGenerator,
	Awaitable,
	Callable,
	Generator,
	Optional,
	Tuple,
	Union,
)


class TestOutcome(Enum):
	"""
	Enumeration representing all possible outcomes of an attempt at running a test.

	Attributes:
									PASS: Represents a passing test outcome - no errors raised, no assertions failed, the test ran to completion.
									FAIL: The test failed in some way - e.g. an assertion failed or an exception was raised.
									SKIP: The test was skipped.
									XFAIL: The test was expected to fail, and it did fail.
									XPASS: The test was expected to fail, however it unexpectedly passed.
									DRYRUN: The test was not executed because the test session was a dry-run.
	"""

	PASS = auto()
	FAIL = auto()
	SKIP = auto()
	XFAIL = auto()	# expected fail
	XPASS = auto()	# unexpected pass
	DRYRUN = auto()	 # tests arent executed during dryruns

	@property
	def display_char(self):
		display_chars = {
			TestOutcome.PASS: ".",
			TestOutcome.FAIL: "F",
			TestOutcome.SKIP: "-",
			TestOutcome.XPASS: "U",
			TestOutcome.XFAIL: "x",
			TestOutcome.DRYRUN: ".",
		}
		assert len(display_chars) == len(TestOutcome)
		return display_chars[self]

	@property
	def display_name(self):
		display_names = {
			TestOutcome.PASS: "Passes",
			TestOutcome.FAIL: "Failures",
			TestOutcome.SKIP: "Skips",
			TestOutcome.XPASS: "Unexpected Passes",
			TestOutcome.XFAIL: "Expected Failures",
			TestOutcome.DRYRUN: "Dry-runs",
		}
		assert len(display_names) == len(TestOutcome)
		return display_names[self]

	@property
	def will_fail_session(self) -> bool:
		return self in {TestOutcome.FAIL, TestOutcome.XPASS}

	@property
	def wont_fail_session(self) -> bool:
		return not self.will_fail_session


@dataclass
class Argument:
	args: list = field(default_factory=list)
	kwargs: dict = field(default_factory=dict)


@dataclass
class Each:
	args: Tuple[Any]

	def __getitem__(self, args):
		return self.args[args]

	def __len__(self):
		return len(self.args)


@dataclass
class Marker:
	name: str
	reason: Optional[str] = None
	when: Union[bool, Callable] = True

	@property
	def active(self):
		try:
			return self.when()
		except TypeError:
			return self.when


@dataclass
class SkipMarker(Marker):
	"""
	Marker that gets attached to a test (via CollectionMetadata) to indicate it should be skipped.
	"""

	name: str = "SKIP"


@dataclass
class ExpectFailMarkup(Marker):
	"""
	Marker that gets attached to a test (via CollectionMetadata) to indicate that we expect it to fail.
	"""

	name: str = "XFAIL"


@dataclass
class CollectionMetadata:
	"""
	Attached to tests and fixtures during the collection phase for later use.
	"""

	marker: Optional[Marker] = None
	comment: Optional[str] = None
	tags: list = field(default_factory=list)
	arguments: list = field(default_factory=list)
	count_of_launchs: int = 1
	is_fixture: bool = False


@dataclass
class Fixture:
	handler: Union[Awaitable, Callable]
	gen: Union[Generator, AsyncGenerator, None] = None
	resolved_val: Any = None

	@property
	def metadata(self):
		return self.handler.pztdmeta
